#include "RobotUnitReader.h"

#include "aron_conversions.h"
#include "Armar6RobotUnitConverter.h"

#include <RobotAPI/libraries/aron/core/navigator/data/container/Dict.h>
#include <RobotAPI/libraries/aron/core/navigator/data/primitive/Long.h>
#include <RobotAPI/libraries/aron/core/navigator/data/primitive/String.h>

#include <RobotAPI/libraries/armem_robot_state/aron/Proprioception.aron.generated.h>

#include <RobotAPI/libraries/RobotUnitDataStreamingReceiver/RobotUnitDataStreamingReceiver.h>

#include <ArmarXCore/core/time/CycleUtil.h>
#include <ArmarXCore/libraries/ArmarXCoreComponentPlugins/DebugObserverComponentPlugin.h>

#include <SimoxUtility/algorithm/string/string_tools.h>

#include <istream>
#include <filesystem>
#include <fstream>


namespace armarx::armem::server::robot_state
{

    std::map<std::string, std::string> RobotUnitReader::readConfig(const std::string& configPath)
    {
        ARMARX_CHECK(std::filesystem::is_regular_file(configPath));

        std::map<std::string, std::string> config;

        std::ifstream in(configPath);
        std::string line;
        while (std::getline(in, line))
        {
            if (line.size() > 0)
            {
                if (simox::alg::starts_with(line, "//"))
                {
                    continue;
                }

                std::vector<std::string> mappings = simox::alg::split(line, ",");
                for (const auto& mapping : mappings)
                {
                    std::vector<std::string> split = simox::alg::split(mapping, "=>");
                    if (split.size() < 2)
                    {
                        ARMARX_WARNING_S << "A mapping in the RobotUnitConfig file is not valid. Ignoring it: " << mapping;
                        continue;
                    }
                    for (unsigned int i = 0; i < split.size() - 1; ++i)
                    {
                        ARMARX_DEBUG_S << "Setting sensor value '" << split[i] << "' to group '" << split[split.size() - 1] << "'.";
                        config.emplace(split[i], split[split.size() - 1]);
                    }
                }
            }
        }

        return config;
    }


    RobotUnitReader::RobotUnitReader() :
        converter(new Armar6RobotUnitConverter())
    {
    }


    void RobotUnitReader::connect(
        armarx::plugins::RobotUnitComponentPlugin& robotUnitPlugin,
        armarx::plugins::DebugObserverComponentPlugin& debugObserverPlugin)
    {
        {
            config.loggingNames.push_back(properties.sensorPrefix);
            receiver = robotUnitPlugin.startDataSatreming(config);
            description = receiver->getDataDescription();
        }
        {
            // Thread-local copy of debug observer helper.
            debugObserver = DebugObserverHelper(
                                Logging::tag.tagName, debugObserverPlugin.getDebugObserver(), true
                            );
        }
    }


    void RobotUnitReader::run(
        float pollFrequency,
        std::queue<RobotUnitData>& dataQueue,
        std::mutex& dataMutex)
    {
        CycleUtil cycle(static_cast<int>(1000.f / pollFrequency));
        while (task and not task->isStopped())
        {
            if (std::optional<RobotUnitData> commit = fetchAndConvertLatestRobotUnitData())
            {
                std::lock_guard g{dataMutex};
                dataQueue.push(std::move(commit.value()));
            }

            if (debugObserver)
            {
                debugObserver->sendDebugObserverBatch();
            }
            cycle.waitForCycleDuration();
        }
    }


    std::optional<RobotUnitData> RobotUnitReader::fetchAndConvertLatestRobotUnitData()
    {
        const std::optional<RobotUnitDataStreaming::TimeStep> data = fetchLatestData();
        if (not data.has_value())
        {
            return std::nullopt;
        }

        ARMARX_DEBUG << "RobotUnitReader: Converting data current timestep to commit";
        auto start = std::chrono::high_resolution_clock::now();

        RobotUnitData result;
        result.proprioception = converter->convert(data.value(), description);
        result.timestamp = Time::microSeconds(data->timestampUSec);;

        auto stop = std::chrono::high_resolution_clock::now();
        auto duration = std::chrono::duration_cast<std::chrono::microseconds>(stop - start);
        ARMARX_DEBUG << "RobotUnitReader: The total time needed to convert the data to commit s: " << duration;

        if (debugObserver)
        {
            debugObserver->setDebugObserverDatafield("RobotUnitReader | t Read+Group [ms]", duration.count() / 1000.f);
        }

        return result;
    }


    std::optional<RobotUnitDataStreaming::TimeStep> RobotUnitReader::fetchLatestData()
    {
        std::deque<RobotUnitDataStreaming::TimeStep>& data = receiver->getDataBuffer();
        if (debugObserver)
        {
            debugObserver->setDebugObserverDatafield("RobotUnitReader | Buffer Size", data.size());
        }
        if (data.empty())
        {
            return std::nullopt;
        }
        else
        {
            RobotUnitDataStreaming::TimeStep currentTimestep = data.back();
            data.clear();
            return currentTimestep;
        }
    }


}

