#pragma once

#include <ArmarXCore/core/logging/Logging.h>
#include <RobotAPI/interface/armem/query.h>


namespace armarx::armem::base::query_proc
{

    /**
     * @brief Base class for memory query processors.
     */
    template <class DataT, class QueryT>
    class BaseQueryProcessorBase
    {
    public:
        using QueryPtrT = ::IceInternal::Handle<QueryT>;
        using QuerySeqT = std::vector<QueryPtrT>;

    public:

        virtual ~BaseQueryProcessorBase() = default;


        DataT process(const QueryT& query, const DataT& data) const
        {
            DataT result = data.copyEmpty();
            if (getTargets(query.targets).count(getTargetType()))
            {
                this->process(result, query, data);
            }
            return result;
        }

        DataT process(const QueryPtrT& query, const DataT& data) const
        {
            return this->process(*query, *data);
        }

        DataT process(const QuerySeqT& queries, const DataT& data) const
        {
            DataT result = data.copyEmpty();
            this->process(result, queries, data);
            return result;
        }

        void process(DataT& result, const QuerySeqT& queries, const DataT& data) const
        {
            if (queries.empty())
            {
                ARMARX_DEBUG << "There are no queries to process.";
                return;
            }

            for (const auto& query : queries)
            {
                if (getTargets(query->targets).count(getTargetType()))
                {
                    this->process(result, *query, data);
                }
            }
        }

        virtual void process(DataT& result, const QueryT& query, const DataT& data) const = 0;


    protected:

        virtual query::data::QueryTarget getTargetType() const = 0;


    private:

        /// If empty, e.g. when receiving queries from python, we use WM as default.
        /// We do it here as (Sl)ice does not support default values for vectors.
        static std::set<query::data::QueryTarget> getTargets(const std::vector<query::data::QueryTarget>& _targets)
        {
            std::set<query::data::QueryTarget> targets(_targets.begin(), _targets.end());
            if (targets.empty())
            {
                ARMARX_DEBUG << "Query has no targets - using WM as default.";
                targets.insert(query::data::QueryTarget::WM);
            }
            return targets;
        }

    };
}
