#pragma once

#include <regex>

#include <ArmarXCore/core/logging/Logging.h>
#include <ArmarXCore/core/exceptions/local/ExpressionException.h>

#include <RobotAPI/interface/armem/query.h>
#include <RobotAPI/libraries/armem/core/error.h>

#include "BaseQueryProcessorBase.h"
#include "ProviderSegmentQueryProcessorBase.h"


namespace armarx::armem::base::query_proc
{

    /**
     * @brief Handles memory queries.
     */
    template <class _CoreSegmentT>
    class CoreSegmentQueryProcessorBase :
        virtual public BaseQueryProcessorBase<_CoreSegmentT, armem::query::data::CoreSegmentQuery>
    {
        using Base = BaseQueryProcessorBase<_CoreSegmentT, armem::query::data::CoreSegmentQuery>;

    public:
        using CoreSegmentT = _CoreSegmentT;
        using ProviderSegmentT = typename _CoreSegmentT::ProviderSegmentT;
        using EntityT = typename ProviderSegmentT::EntityT;
        using EntitySnapshotT = typename EntityT::EntitySnapshotT;

        using Base::process;
        virtual void process(_CoreSegmentT& result,
                             const armem::query::data::CoreSegmentQuery& query,
                             const _CoreSegmentT& coreSegment) const override
        {
            if (auto q = dynamic_cast<const armem::query::data::core::All*>(&query))
            {
                process(result, *q, coreSegment);
            }
            else if (auto q = dynamic_cast<const armem::query::data::core::Single*>(&query))
            {
                process(result, *q, coreSegment);
            }
            else if (auto q = dynamic_cast<const armem::query::data::core::Regex*>(&query))
            {
                process(result, *q, coreSegment);
            }
            else
            {
                throw armem::error::UnknownQueryType("core segment", query);
            }
        }

        void process(_CoreSegmentT& result,
                     const armem::query::data::core::All& query,
                     const _CoreSegmentT& coreSegment) const
        {
            for (const auto& [name, providerSegment] : coreSegment.providerSegments())
            {
                result.addProviderSegment(providerSegmentProcessorProcess(query.providerSegmentQueries, providerSegment));
            }
        }

        void process(_CoreSegmentT& result,
                     const armem::query::data::core::Single& query,
                     const _CoreSegmentT& coreSegment) const
        {
            try
            {
                const ProviderSegmentT& providerSegment = coreSegment.getProviderSegment(query.providerSegmentName);
                result.addProviderSegment(providerSegmentProcessorProcess(query.providerSegmentQueries, providerSegment));
            }
            catch (const error::MissingEntry&)
            {
                // Leave empty.
            }
        }

        void process(_CoreSegmentT& result,
                     const armem::query::data::core::Regex& query,
                     const _CoreSegmentT& coreSegment) const
        {
            std::regex regex(query.providerSegmentNameRegex);
            for (const auto& [name, providerSegment] : coreSegment.providerSegments())
            {
                if (std::regex_search(providerSegment.name(), regex))
                {
                    result.addProviderSegment(providerSegmentProcessorProcess(query.providerSegmentQueries, providerSegment));
                }
            }
        }

    protected:
        virtual ProviderSegmentT providerSegmentProcessorProcess(const armem::query::data::ProviderSegmentQuerySeq& a, const ProviderSegmentT& o) const = 0;
    };
}
