#pragma once

#include <regex>

#include <RobotAPI/interface/armem/query.h>

#include "BaseQueryProcessorBase.h"
#include "CoreSegmentQueryProcessorBase.h"


#include <ArmarXCore/core/logging/Logging.h>
#include <ArmarXCore/core/exceptions/local/ExpressionException.h>

#include <RobotAPI/libraries/armem/core/error.h>


namespace armarx::armem::base::query_proc
{
    /**
     * @brief Handles memory queries.
     */
    template <class _MemoryT>
    class MemoryQueryProcessorBase :
        virtual public BaseQueryProcessorBase<_MemoryT, armem::query::data::MemoryQuery>
    {
        using Base = BaseQueryProcessorBase<_MemoryT, armem::query::data::MemoryQuery>;

    public:
        using MemoryT = _MemoryT;
        using CoreSegmentT = typename MemoryT::CoreSegmentT;
        using ProviderSegmentT = typename CoreSegmentT::ProviderSegmentT;
        using EntityT = typename ProviderSegmentT::EntityT;
        using EntitySnapshotT = typename EntityT::EntitySnapshotT;


        using Base::process;
        _MemoryT process(const armem::query::data::Input& input, const _MemoryT& memory) const
        {
            return this->process(input.memoryQueries, memory);
        }

        void process(_MemoryT& result,
                     const armem::query::data::MemoryQuery& query,
                     const _MemoryT& memory) const override
        {
            if (auto q = dynamic_cast<const armem::query::data::memory::All*>(&query))
            {
                process(result, *q, memory);
            }
            else if (auto q = dynamic_cast<const armem::query::data::memory::Single*>(&query))
            {
                process(result, *q, memory);
            }
            else if (auto q = dynamic_cast<const armem::query::data::memory::Regex*>(&query))
            {
                process(result, *q, memory);
            }
            else
            {
                throw armem::error::UnknownQueryType("memory segment", query);
            }
        }

        void process(_MemoryT& result,
                     const armem::query::data::memory::All& query,
                     const _MemoryT& memory) const
        {
            for (const auto& [name, coreSegment] : memory.coreSegments())
            {
                result.addCoreSegment(coreSegmentProcessorProcess(query.coreSegmentQueries, coreSegment));
            }
        }

        void process(_MemoryT& result,
                     const armem::query::data::memory::Single& query,
                     const _MemoryT& memory) const
        {
            try
            {
                const CoreSegmentT& coreSegment = memory.getCoreSegment(query.coreSegmentName);
                result.addCoreSegment(coreSegmentProcessorProcess(query.coreSegmentQueries, coreSegment));
            }
            catch (const error::MissingEntry&)
            {
            }
        }

        void process(_MemoryT& result,
                     const armem::query::data::memory::Regex& query,
                     const _MemoryT& memory) const
        {
            std::regex regex(query.coreSegmentNameRegex);
            for (const auto& [name, coreSegment] : memory.coreSegments())
            {
                if (std::regex_search(coreSegment.name(), regex))
                {
                    result.addCoreSegment(coreSegmentProcessorProcess(query.coreSegmentQueries, coreSegment));
                }
            }
        }

    protected:
        virtual CoreSegmentT coreSegmentProcessorProcess(const armem::query::data::CoreSegmentQuerySeq& a, const CoreSegmentT& o) const = 0;
    };
}
