#include "MemoryViewer.h"

#include <RobotAPI/libraries/armem/core/workingmemory/ice_conversions.h>
#include <RobotAPI/libraries/armem_gui/gui_utils.h>

#include <RobotAPI/libraries/armem/core/diskmemory/Memory.h>

#include <RobotAPI/libraries/armem/server/query_proc/diskmemory/MemoryQueryProcessor.h>
#include <RobotAPI/libraries/armem/server/query_proc/workingmemory/MemoryQueryProcessor.h>

#include <ArmarXGui/libraries/SimpleConfigDialog/SimpleConfigDialog.h>

#include <ArmarXCore/core/ManagedIceObject.h>
#include <ArmarXCore/core/time/TimeUtil.h>
#include <ArmarXCore/observers/variant/Variant.h>

#include <SimoxUtility/algorithm/get_map_keys_values.h>

#include <QBoxLayout>
#include <QCheckBox>
#include <QDialog>
#include <QGroupBox>
#include <QMenu>
#include <QLabel>
#include <QLayout>
#include <QSettings>

#include <filesystem>

namespace armarx::armem::gui
{
    MemoryViewer::MemoryViewer(
        QBoxLayout* updateWidgetLayout,
        QBoxLayout* memoryControlWidgetLayout,
        QGroupBox* memoryGroupBox, QLayout* memoryGroupBoxParentLayout,
        QGroupBox* instanceGroupBox, QLayout* instanceGroupBoxParentLayout,
        QLabel* statusLabel
    )
    {
        Logging::setTag("MemoryViewer");

        this->statusLabel = statusLabel;
        this->statusLabel->clear();

        statusLabel->setContextMenuPolicy(Qt::CustomContextMenu);
        connect(statusLabel, &QLabel::customContextMenuRequested, [statusLabel](const QPoint & pos)
        {
            QMenu menu(statusLabel);
            menu.addAction("Clear status", [statusLabel]()
            {
                statusLabel->clear();
            });
            menu.exec(statusLabel->mapToGlobal(pos));
        });


        // Update timer
        this->updateWidgetLayout = updateWidgetLayout;
        updateWidget = new armem::gui::PeriodicUpdateWidget(2.0, 60);
        updateWidgetLayout->insertWidget(0, updateWidget);

        // LTM Control
        if (memoryControlWidgetLayout)
        {
            this->memoryControlWidgetLayout = memoryControlWidgetLayout;
            memoryControlWidget = new armem::gui::MemoryControlWidget();
            memoryControlWidgetLayout->addWidget(memoryControlWidget);
        }

        // Memory View
        memoryGroup = new armem::gui::MemoryGroupBox();
        armarx::gui::replaceWidget(memoryGroupBox, memoryGroup, memoryGroupBoxParentLayout);
        ARMARX_CHECK_NULL(memoryGroupBox);

        // Instance View
        this->instanceGroup = new armem::gui::InstanceGroupBox();
        armarx::gui::replaceWidget(instanceGroupBox, instanceGroup, instanceGroupBoxParentLayout);
        this->instanceGroup->setStatusLabel(statusLabel);
        ARMARX_CHECK_NULL(instanceGroupBox);

        // Connections
        //connect(this, &This::connected, this, &This::updateMemory);

        connect(memoryControlWidget, &armem::gui::MemoryControlWidget::storeOnDisk, this, &This::storeOnDisk);
        connect(memoryControlWidget, &armem::gui::MemoryControlWidget::storeInLTM, this, &This::storeInLTM);

        connect(this, &This::connected, this, &This::updateMemories);
        connect(updateWidget, &armem::gui::PeriodicUpdateWidget::update, this, &This::updateMemories);

        connect(this, &This::memoryDataChanged, this, &This::updateMemoryTree);
        connect(memoryGroup->tree(), &armem::gui::MemoryTreeWidget::selectedItemChanged, this, &This::updateInstanceTree);

        connect(memoryGroup->tree(), &armem::gui::MemoryTreeWidget::updated, this, &This::memoryTreeUpdated);
        connect(instanceGroup, &armem::gui::InstanceGroupBox::viewUpdated, this, &This::instanceTreeUpdated);
        connect(instanceGroup->view, &armem::gui::InstanceView::memoryIdResolutionRequested, this, &This::resolveMemoryID);
    }

    void MemoryViewer::setLogTag(const std::string& tag)
    {
        Logging::setTag(tag);
    }

    void MemoryViewer::onInit(ManagedIceObject& component)
    {
        if (mnsName.size() > 0)
        {
            component.usingProxy(mnsName);
        }
        if (debugObserverName.size() > 0)
        {
            component.usingProxy(debugObserverName);
        }

        emit initialized();
    }

    void MemoryViewer::onConnect(ManagedIceObject& component)
    {
        if (!mnsName.empty())
        {
            armem::mns::MemoryNameSystemInterfacePrx mnsProxy;
            component.getProxy(mnsProxy, mnsName);
            mns = client::MemoryNameSystem(mnsProxy);

            const bool update = true;
            memoryReaders = mns.getAllReaders(update);
        }
        // DebugObserver is optional (check for null on every call)
        if (!debugObserverName.empty())
        {
            component.getProxy(debugObserver, debugObserverName, false, "", false);
        }
        updateWidget->startTimerIfEnabled();

        emit connected();
    }

    void MemoryViewer::onDisconnect(ManagedIceObject&)
    {
        updateWidget->stopTimer();

        emit disconnected();
    }

    const armem::wm::Memory* MemoryViewer::getSingleMemoryData(const std::string& memoryName)
    {
        auto it = memoryData.find(memoryName);
        if (it == memoryData.end())
        {
            std::stringstream ss;
            ss << "Memory name '" << memoryName << "' is unknown. Known are: "
               << simox::alg::get_keys(memoryData);
            statusLabel->setText(QString::fromStdString(ss.str()));
            return nullptr;
        }
        /*else if (not it->second.has_value())
        {
            return nullptr;
        }*/
        else
        {
            //return &it->second.value();
            return &it->second;
        }
    }

    void MemoryViewer::storeInLTM()
    {
        TIMING_START(MemoryStore);

        for (auto& [name, reader] : memoryReaders)
        {
            data::StoreInput input;
            input.query = memoryGroup->queryWidget()->queryInput().toIce();
            reader.readAndStore(input);
        }

        TIMING_END_STREAM(MemoryStore, ARMARX_VERBOSE);
    }

    void MemoryViewer::storeOnDisk()
    {
        TIMING_START(MemoryExport);
        QString qs = memoryControlWidget->getEnteredPath();
        std::string utf8_text = qs.toUtf8().constData();

        if (not utf8_text.empty())
        {
            ARMARX_IMPORTANT << "Exporting all memories at '" << utf8_text << "'.";

            std::filesystem::path p(utf8_text);
            if (std::filesystem::is_regular_file(p))
            {
                ARMARX_WARNING << "Could not export a memory at '" << utf8_text << "'. Skipping export.";
                return;
            }

            std::filesystem::create_directories(p);
            for (auto& [name, reader] : memoryReaders)
            {
                armem::client::QueryInput input = memoryGroup->queryWidget()->queryInput();
                armem::client::QueryResult result = reader.query(input);

                armem::d_ltm::Memory dMem(name);
                dMem.reload(p / name);
                dMem.append(result.memory);
            }
        }
        else
        {
            ARMARX_WARNING << "The path is empty. Could not export the memory in nirvana.";
        }

        TIMING_END_STREAM(MemoryExport, ARMARX_VERBOSE);
    }

    void MemoryViewer::updateMemories()
    {
        memoryReaders.clear();
        memoryData.clear();

        memoryReaders = mns.getAllReaders(true);

        bool dataChanged = false;

        QString qs = memoryControlWidget->getEnteredPath();
        std::string utf8_text = qs.toUtf8().constData();

        std::filesystem::path p(utf8_text);

        // first check if the local file system should be queried
        if (memoryGroup->queryWidget()->alsoQueryLocalDisk())
        {
            if (std::filesystem::is_directory(p))
            {
                for (const auto& d : std::filesystem::directory_iterator(p))
                {
                    if (d.is_directory())
                    {
                        std::string k = d.path().filename();
                        armem::d_ltm::Memory dMem(k);
                        dMem.reload(p / k);

                        armem::client::QueryInput input = memoryGroup->queryWidget()->queryInput();
                        input.addQueryTargetToAll(armem::query::data::QueryTarget::Disk);

                        armem::d_ltm::query_proc::MemoryQueryProcessor d_ltm_processor;
                        dMem = d_ltm_processor.process(input.toIce(), dMem);

                        wm::Memory converted = dMem.convert();
                        memoryData[k] = std::move(converted);
                        dataChanged = true;
                    }
                }
            }
            else
            {
                ARMARX_WARNING << "Could not import a memory from '" << utf8_text << "'. Skipping import.";
            }
        }

        armem::client::QueryInput input = memoryGroup->queryWidget()->queryInput();
        for (auto& [name, reader] : memoryReaders)
        {
            TIMING_START(MemoryQuery);
            {
                armem::client::QueryResult result = reader.query(input);
                if (result)
                {
                    if (result.memory.hasData())
                    {
                        if (const auto& it = memoryData.find(name); it != memoryData.end())
                        {
                            result.memory.append(it->second);

                            // requery (e.g. to get only the last n values instead of the last n from disk and the last n from wm)
                            armem::wm::query_proc::MemoryQueryProcessor wm_processor;
                            result.memory = wm_processor.process(input.toIce(), result.memory);

                            if (!result.memory.hasData())
                            {
                                ARMARX_ERROR << "A memory which had data before lost data. This indicates that there is something wrong.";
                            }
                        }

                        dataChanged = true;
                        memoryData[name] = std::move(result.memory);
                    }
                    else
                    {
                        ARMARX_INFO << "The memory " << name << " has no data after querying.";
                    }
                }
                else
                {
                    ARMARX_WARNING << "A query for memory '" << name << "' produced an error: " << result.errorMessage;
                    if (statusLabel)
                    {
                        statusLabel->setText(QString::fromStdString(result.errorMessage));
                    }
                }
            }
            TIMING_END_STREAM(MemoryQuery, ARMARX_VERBOSE);

            if (debugObserver)
            {
                debugObserver->setDebugDatafield(Logging::tag.tagName, "Memory Query [ms]", new Variant(MemoryQuery.toMilliSecondsDouble()));
            }
        }

        if (dataChanged)
        {
            emit memoryDataChanged();
        }
        else
        {
            if (statusLabel)
            {
                statusLabel->setText("No query result.");
            }
        }
    }


    void MemoryViewer::updateInstanceTree(const armem::MemoryID& selectedID)
    {
        const armem::wm::Memory* data = getSingleMemoryData(selectedID.memoryName);
        if (data)
        {
            if (!selectedID.hasEntityName())
            {
                return;
            }
            armem::MemoryID id = selectedID;
            const armem::wm::EntitySnapshot* snapshot = nullptr;
            if (!id.hasTimestamp())
            {
                const armem::wm::Entity& entity = data->getEntity(id);
                if (entity.empty())
                {
                    return;
                }
                snapshot = &entity.getLatestSnapshot();
                id.timestamp = snapshot->time();
            }
            if (!id.hasInstanceIndex())
            {
                if (!snapshot)
                {
                    try
                    {
                        snapshot = &data->getEntitySnapshot(id);
                    }
                    catch (const armem::error::ArMemError& e)
                    {
                        if (statusLabel)
                        {
                            statusLabel->setText(e.what());
                        }
                    }
                }
                if (snapshot && snapshot->size() > 0)
                {
                    id.instanceIndex = 0;
                }
            }
            if (id.hasInstanceIndex())
            {
                instanceGroup->view->update(id, *data);
            }
        }
    }

    void MemoryViewer::resolveMemoryID(const MemoryID& id)
    {
        // ARMARX_IMPORTANT << "Resolving memory ID: " << id;

        auto handleError = [this](const std::string & msg)
        {
            statusLabel->setText(QString::fromStdString(msg));
            ARMARX_WARNING << msg;
        };

        if (id.memoryName.empty())
        {
            handleError("Memory name is empty.");
        }

        aron::typenavigator::ObjectNavigatorPtr segmentType;
        std::optional<wm::EntityInstance> instance;
        try
        {
            const wm::Memory* data = getSingleMemoryData(id.memoryName);
            if (data)
            {
                segmentType = data->getProviderSegment(id).aronType();

                if (id.hasInstanceIndex())
                {
                    instance = data->getEntityInstance(id);
                }
                else if (id.hasTimestamp())
                {
                    instance = data->getEntitySnapshot(id).getInstance(0);
                }
                else
                {
                    instance = data->getEntity(id).getLatestSnapshot().getInstance(0);
                }
            }
        }
        catch (const armem::error::ArMemError& e)
        {
            // May be handled by remote lookup
            (void) e;
        }

        if (!instance)
        {
            try
            {
                // Resolve remotely (may still fail, returns an optional).
                instance = mns.resolveEntityInstance(id);
            }
            catch (const armem::error::ArMemError& e)
            {
                ARMARX_WARNING << e.what();
                statusLabel->setText(e.what());
            }
        }

        if (instance)
        {
            instanceGroup->view->addInstanceView(*instance, segmentType);
        }
        else
        {
            // ToDo: Propagate error back to highlight selected entry in red
        }
    }

    void MemoryViewer::updateMemoryTree()
    {
        std::map<std::string, const armem::wm::Memory*> convMap;
        for (auto& [name, data] : memoryData)
        {
            /*if (data.has_value())
            {
                convMap[name] = &data.value();
            }*/
            convMap[name] = &data;
        }

        if (convMap.empty())
        {
            return;
        }

        TIMING_START(GuiUpdate);
        memoryGroup->tree()->update(convMap);
        TIMING_END_STREAM(GuiUpdate, ARMARX_VERBOSE);

        if (debugObserver)
        {
            debugObserver->setDebugDatafield(Logging::tag.tagName, "GUI Update [ms]", new Variant(GuiUpdate.toMilliSecondsDouble()));
        }
    }

    const static std::string CONFIG_KEY_MEMORY = "MemoryViewer.MemoryNameSystem";
    const static std::string CONFIG_KEY_DEBUG_OBSERVER = "MemoryViewer.DebugObserverName";

    void MemoryViewer::loadSettings(QSettings* settings)
    {
        mnsName = settings->value(QString::fromStdString(CONFIG_KEY_MEMORY), "MemoryNameSystem").toString().toStdString();
        debugObserverName = settings->value(QString::fromStdString(CONFIG_KEY_DEBUG_OBSERVER), "DebugObserver").toString().toStdString();
    }
    void MemoryViewer::saveSettings(QSettings* settings)
    {
        settings->setValue(QString::fromStdString(CONFIG_KEY_MEMORY), QString::fromStdString(mnsName));
        settings->setValue(QString::fromStdString(CONFIG_KEY_DEBUG_OBSERVER), QString::fromStdString(debugObserverName));
    }

    void MemoryViewer::writeConfigDialog(SimpleConfigDialog* dialog)
    {
        dialog->addProxyFinder<armarx::armem::mns::MemoryNameSystemInterfacePrx>({CONFIG_KEY_MEMORY, "MemoryNameSystem", "*"});
        dialog->addProxyFinder<armarx::DebugObserverInterfacePrx>({CONFIG_KEY_DEBUG_OBSERVER, "Debug Observer", "DebugObserver"});
    }

    void MemoryViewer::readConfigDialog(SimpleConfigDialog* dialog)
    {
        mnsName = dialog->getProxyName(CONFIG_KEY_MEMORY);
        if (mnsName.empty())
        {
            mnsName = "MemoryNameSystem";
        }
        debugObserverName = dialog->getProxyName(CONFIG_KEY_DEBUG_OBSERVER);
    }

}
