/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2012-2016, High Performance Humanoid Technologies (H2T),
 * Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Fabian Peller-Konrad (fabian dot peller-konrad at kit dot edu)
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#pragma once

// STD/STL
#include <memory>
#include <numeric>

// ArmarX
#include <RobotAPI/libraries/aron/core/Exception.h>
#include <RobotAPI/libraries/aron/core/navigator/data/AllNavigators.h>
#include <RobotAPI/libraries/aron/core/navigator/type/AllNavigators.h>

namespace armarx::aron
{
    class Randomizer
    {
    public:
        Randomizer()
        {
            initializeSelf();
        };

        typenavigator::NavigatorPtr generateRandomType(bool allowMaybeType = false, bool firstObject = false) const
        {
            type::Descriptor nextType = type::Descriptor::eObject;
            if (!firstObject)
            {
                nextType = getRandomElement(type::ALL_ARON_TYPES);
            }

            type::Maybe nextMaybeType = type::Maybe::eNone;
            if (allowMaybeType && !firstObject)
            {
                nextMaybeType = getRandomElement(type::ALL_MAYBE_TYPES);
            }
            (void) nextMaybeType;  // Unused

            switch (nextType)
            {
                case type::Descriptor::eObject:
                {
                    auto t = std::make_shared<typenavigator::ObjectNavigator>();
                    std::string objectName = generateRandomWord();
                    t->setObjectName(objectName);

                    int members = generateRandom(4, 2);
                    std::set<std::string> usedKeys = {objectName};
                    for (int i = 0; i < members; ++i)
                    {
                        std::string key = generateRandomWord(usedKeys);
                        usedKeys.insert(key);

                        typenavigator::NavigatorPtr m = generateRandomType(false);
                        t->addMemberType(key, m);
                    }
                    return t;
                }
                case type::Descriptor::eIntEnum:
                {
                    auto t = std::make_shared<typenavigator::IntEnumNavigator>();
                    std::string intEnumName = generateRandomWord();
                    t->setEnumName(intEnumName);

                    std::set<std::string> usedKeys = {intEnumName};
                    for (int i = 0; i < generateRandom(5, 2); ++i)
                    {
                        std::string key = generateRandomWord(usedKeys);
                        usedKeys.insert(key);

                        t->addAcceptedValue(key, i);
                    }
                    return t;
                }
                case type::Descriptor::eDict:
                {
                    auto t = std::make_shared<typenavigator::DictNavigator>();
                    typenavigator::NavigatorPtr a = generateRandomType(false);

                    t->setAcceptedType(a);
                    return t;
                }
                case type::Descriptor::eTuple:
                {
                    auto t = std::make_shared<typenavigator::TupleNavigator>();

                    int members = generateRandom(4, 2);
                    for (int i = 0; i < members; ++i)
                    {
                        typenavigator::NavigatorPtr m = generateRandomType(false);
                        t->addAcceptedType(m);
                    }
                    return t;
                }
                case type::Descriptor::eList:
                {
                    auto t = std::make_shared<typenavigator::ListNavigator>();
                    typenavigator::NavigatorPtr a = generateRandomType(false);

                    t->setAcceptedType(a);
                    return t;
                }
                case type::Descriptor::ePair:
                {
                    auto t = std::make_shared<typenavigator::PairNavigator>();
                    typenavigator::NavigatorPtr a = generateRandomType(false);
                    typenavigator::NavigatorPtr b = generateRandomType(false);

                    t->setFirstAcceptedType(a);
                    t->setSecondAcceptedType(b);
                    return t;
                }
                case type::Descriptor::eNDArray:
                {
                    // TODO (fabian.peller)
                    // fall through EigenMatrix
                }
                case type::Descriptor::eEigenMatrix:
                {
                    auto t = std::make_shared<typenavigator::EigenMatrixNavigator>();
                    auto type = getRandomKey(t->ACCEPTED_TYPES);
                    t->setTypename(type);
                    t->setRows(generateRandom(10, 1));
                    t->setRows(generateRandom(10, 1));
                    return t;
                }
                case type::Descriptor::eEigenQuaternion:
                {
                    auto t = std::make_shared<typenavigator::EigenQuaternionNavigator>();
                    auto type = getRandomKey(t->ACCEPTED_TYPES);
                    t->setTypename(type);
                    return t;
                }
                case type::Descriptor::eIVTCByteImage:
                {
                    auto t = std::make_shared<typenavigator::IVTCByteImageNavigator>();
                    auto type = getRandomKey(t->ACCEPTED_TYPES);
                    return t;
                }
                case type::Descriptor::eOpenCVMat:
                {
                    auto t = std::make_shared<typenavigator::OpenCVMatNavigator>();
                    auto type = getRandomKey(t->ACCEPTED_TYPES);
                    return t;
                }
                case type::Descriptor::ePCLPointCloud:
                {
                    auto t = std::make_shared<typenavigator::PCLPointCloudNavigator>();
                    auto type = getRandomKey(t->ACCEPTED_TYPES);
                    t->setTypename(type);
                    return t;
                }
                case type::Descriptor::ePosition:
                {
                    auto t = std::make_shared<typenavigator::PositionNavigator>();
                    return t;
                }
                case type::Descriptor::eOrientation:
                {
                    auto t = std::make_shared<typenavigator::PositionNavigator>();
                    return t;
                }
                case type::Descriptor::ePose:
                {
                    auto t = std::make_shared<typenavigator::PositionNavigator>();
                    return t;
                }
                case type::Descriptor::eInt:
                {
                    auto t = std::make_shared<typenavigator::IntNavigator>();
                    return t;
                }
                case type::Descriptor::eLong:
                {
                    auto t = std::make_shared<typenavigator::LongNavigator>();
                    return t;
                }
                case type::Descriptor::eFloat:
                {
                    auto t = std::make_shared<typenavigator::FloatNavigator>();
                    return t;
                }
                case type::Descriptor::eDouble:
                {
                    auto t = std::make_shared<typenavigator::DoubleNavigator>();
                    return t;
                }
                case type::Descriptor::eString:
                {
                    auto t = std::make_shared<typenavigator::StringNavigator>();
                    return t;
                }
                case type::Descriptor::eBool:
                {
                    auto t = std::make_shared<typenavigator::BoolNavigator>();
                    return t;
                }
                case type::Descriptor::eTime:
                {
                    auto t = std::make_shared<typenavigator::TimeNavigator>();
                    return t;
                }
                case type::Descriptor::eUnknown:
                {
                    throw error::DescriptorNotValidException("Randomizer", "generateRandomType", "No valid type found!", nextType);
                }
            }
        }

        datanavigator::NavigatorPtr generateEmptyAronDataFromType(const typenavigator::NavigatorPtr& type, bool ignore_maybe = false) const
        {
            if (type->getMaybe() != type::Maybe::eNone && !ignore_maybe)
            {
                if (fiftyPercentChance())
                {
                    return nullptr;
                }
            }

            const type::Descriptor desc = type->getDescriptor();
            switch (desc)
            {
                // In an object, we do not want to edit the keys.
                case type::Descriptor::eObject:
                {
                    typenavigator::ObjectNavigatorPtr t = typenavigator::ObjectNavigator::DynamicCastAndCheck(type);
                    auto d = std::make_shared<datanavigator::DictNavigator>();
                    for (const auto& [k, tt] : t->getMemberTypes())
                    {
                        d->addElement(k, generateEmptyAronDataFromType(tt));
                    }
                    return d;
                }

                case type::Descriptor::eIntEnum:
                {
                    auto t = typenavigator::IntEnumNavigator::DynamicCastAndCheck(type);
                    auto d = std::make_shared<datanavigator::IntNavigator>();
                    return d;
                }

                // here all totally random
                case type::Descriptor::eDict:
                {
                    auto t = typenavigator::DictNavigator::DynamicCastAndCheck(type);
                    return datanavigator::NavigatorPtr(new datanavigator::DictNavigator());
                }

#define RUN_ARON_MACRO(upperType, lowerType, capsType) \
case type::Descriptor::e##upperType: \
{ \
    auto t = typenavigator::upperType##Navigator::DynamicCastAndCheck(type); \
    return datanavigator::NavigatorPtr(new datanavigator::ListNavigator()); \
}

                HANDLE_LIST_SERIALIZER_TYPES
#undef RUN_ARON_MACRO

#define RUN_ARON_MACRO(upperType, lowerType, capsType) \
case type::Descriptor::e##upperType: \
{ \
    auto t = typenavigator::upperType##Navigator::DynamicCastAndCheck(type); \
    return datanavigator::NDArrayNavigatorPtr(new datanavigator::NDArrayNavigator()); \
}

                HANDLE_NDARRAY_TYPES
#undef RUN_ARON_MACRO

#define RUN_ARON_MACRO(upperType, lowerType, capsType, upperData, lowerData, capsData) \
case type::Descriptor::e##upperType: \
{ \
    auto t = typenavigator::upperType##Navigator::DynamicCastAndCheck(type); \
    return datanavigator::NavigatorPtr(new datanavigator::upperData##Navigator()); \
}

                HANDLE_PRIMITIVE_CORRESPONDING
#undef RUN_ARON_MACRO
                case type::Descriptor::eUnknown:
                {
                    throw error::DescriptorNotValidException("Randomizer", "generateAronDataFromType", "No valid type found!", desc);
                }
            }
            // end of non void....... dont know why
            return nullptr;
        }

        void initializeRandomly(datanavigator::NavigatorPtr& data, const typenavigator::NavigatorPtr& type) const
        {
            data = generateEmptyAronDataFromType(type);
            if (!data)
            {
                if (type->getMaybe() == type::Maybe::eNone)
                {
                    throw error::AronException("Randomizer", "initializeRandomly", "The data is null but the type is not a maybe type. This is not valid.");
                }
                return;
            }

            // Containers
            type::Descriptor desc = type->getDescriptor();
            switch (desc)
            {

#define RUN_ARON_MACRO(upperType, lowerType, capsType, upperData, lowerData, capsData) \
case type::Descriptor::e##upperType: \
{ \
    typenavigator::upperType##NavigatorPtr t = typenavigator::upperType##Navigator::DynamicCastAndCheck(type); \
    datanavigator::upperData##NavigatorPtr d = datanavigator::upperData##Navigator::DynamicCastAndCheck(data); \
    initializeRandomly(d, t); \
    break; \
}

                HANDLE_ALL_CORRESPONDING
#undef RUN_ARON_MACRO

                case type::Descriptor::eUnknown:
                {
                    throw error::DescriptorNotValidException("Randomizer", "initializeRandomly", "No valid type found!", desc);
                }
            }
        }

        // generate i in [min, max)
        int generateRandom(int max, int min) const
        {
            if (max < min)
            {
                throw error::IndexNotValidException("AronRandomizer", "generateRandom", "The max value is lower than the min value", max, min);
            }
            int random = (std::rand() % (max - min)) + min;
            return random;
        }

        template <class T>
        T getRandomElement(const std::vector<T>& vec) const
        {
            int i = generateRandom(vec.size(), 0);
            return vec.at(i);
        }

        template <class T>
        std::string getRandomKey(const std::map<std::string, T>& m) const
        {
            std::vector<std::string> keys;
            for (const auto [k, _] : m)
            {
                keys.push_back(k);
            }
            return getRandomElement(keys);
        }

        bool fiftyPercentChance() const
        {
            return generateRandom(2, 0);
        }

        std::string generateRandomWord(const std::set<std::string>& usedKeys = {}) const
        {
            // https://randomwordgenerator.com/
            // script to convert random words into string:
            // jQuery("#result li i").remove(); var str = ""; jQuery("#result li").each(function() {str += jQuery(this).html() + ", " }); console.log(str);
            std::vector<string> words =
            {
                "jaw", "river", "bow", "profession", "heat", "interference", "slave", "population", "deport", "redeem", "digress", "penny", "cousin", "beef", "Bible", "fuss",
                "urgency", "tasty", "innovation", "upset", "gold", "day", "remunerate", "strain", "district", "panel", "choke", "rack", "parameter", "despair", "extort", "country",
                "hesitate", "record", "brand", "confusion", "discreet", "accept", "lifestyle", "option", "corn", "charity", "miss", "viable", "glance", "norm", "meet", "bird",
                "ribbon", "guideline"
            };

            int i = generateRandom(words.size(), 0);
            std::string key = words.at(i);

            while (usedKeys.count(key) > 0)
            {
                key = generateRandomWord(usedKeys);
            }

            return key;
        }

        std::vector<unsigned char> generateRandomBlob(unsigned int size) const
        {
            std::vector<unsigned char> new_blob(size, 0);
            for (unsigned int i = 0; i < size; ++i)
            {
                new_blob[i] = (generateRandom(128, 0));
            }
            return new_blob;
        }

        void initializeRandomly(datanavigator::DictNavigatorPtr& data, const typenavigator::ObjectNavigatorPtr& type) const
        {
            for (auto& [key, nextData] : data->getElements())
            {
                initializeRandomly(nextData, type->getMemberTypes()[key]);
            }
        }

        void initializeRandomly(datanavigator::DictNavigatorPtr& data, const typenavigator::DictNavigatorPtr& type) const
        {
            data->clear();
            int numElements = generateRandom(5, 1);
            std::set<std::string> usedKeys;
            for (int i = 0; i < numElements; ++i)
            {
                std::string key = generateRandomWord(usedKeys);
                usedKeys.insert(key);
                datanavigator::NavigatorPtr newData = generateEmptyAronDataFromType(type->getAcceptedType());
                initializeRandomly(newData, type->getAcceptedType());
                data->addElement(key, newData);
            }
        }

        void initializeRandomly(datanavigator::ListNavigatorPtr& data, const typenavigator::TupleNavigatorPtr& type) const
        {
            unsigned int i = 0;
            for (auto& nextData : data->getElements())
            {
                initializeRandomly(nextData, type->getAcceptedType(i++));
            }
        }

        void initializeRandomly(datanavigator::ListNavigatorPtr& data, const typenavigator::PairNavigatorPtr& type) const
        {
            datanavigator::NavigatorPtr first = data->getElement(0);
            datanavigator::NavigatorPtr second = data->getElement(1);
            initializeRandomly(first, type->getFirstAcceptedType());
            initializeRandomly(second, type->getSecondAcceptedType());
        }

        void initializeRandomly(datanavigator::ListNavigatorPtr& data, const typenavigator::ListNavigatorPtr& type) const
        {
            data->clear();
            int numElements = generateRandom(5, 1);
            for (int i = 0; i < numElements; ++i)
            {
                datanavigator::NavigatorPtr newData = generateEmptyAronDataFromType(type->getAcceptedType());
                initializeRandomly(newData, type->getAcceptedType());
                data->addElement(newData);
            }
        }

        void initializeRandomly(datanavigator::NDArrayNavigatorPtr& data, const typenavigator::NavigatorPtr& type) const
        {
            std::vector<int> dims = data->getDimensions();
            int bytes = std::accumulate(std::begin(dims), std::end(dims), 1, std::multiplies<int>());
            std::vector<unsigned char> blob = generateRandomBlob(bytes);
            data->setData(bytes, blob.data());
        }

        void initializeRandomly(datanavigator::IntNavigatorPtr& data, const typenavigator::NavigatorPtr& type) const
        {
            data->setValue(generateRandom(1000, -1000));
        }

        void initializeRandomly(datanavigator::LongNavigatorPtr& data, const typenavigator::NavigatorPtr& type) const
        {
            data->setValue(generateRandom(1000, -1000));
        }

        void initializeRandomly(datanavigator::FloatNavigatorPtr& data, const typenavigator::NavigatorPtr& type) const
        {
            data->setValue(generateRandom(1000, -1000));
        }

        void initializeRandomly(datanavigator::DoubleNavigatorPtr& data, const typenavigator::NavigatorPtr& type) const
        {
            data->setValue(generateRandom(1000, -1000));
        }

        void initializeRandomly(datanavigator::BoolNavigatorPtr& data, const typenavigator::NavigatorPtr& type) const
        {
            data->setValue(fiftyPercentChance());
        }

        void initializeRandomly(datanavigator::StringNavigatorPtr& data, const typenavigator::NavigatorPtr& type) const
        {
            data->setValue(generateRandomWord());
        }

    private:
        void initializeSelf() const
        {
            std::srand(std::time(nullptr));
        }
    };
}
