/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2012-2016, High Performance Humanoid Technologies (H2T),
 * Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Fabian Peller-Konrad (fabian dot peller-konrad at kit dot edu)
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

// Header
#include "OpenCVMat.h"

namespace armarx::aron::cppserializer::serializer
{

    const std::map<std::string, std::pair<std::string, int>> OpenCVMatSerializer::ACCEPTED_TYPES =
    {
        {"CV_8U", {"CV_8U", 1}},
        {"CV_8S", {"CV_8S", 1}},
        {"CV_16U", {"CV_16U", 2}},
        {"CV_16S", {"CV_16S", 2}},
        {"CV_32S", {"CV_32S", 4}},
        {"CV_32F", {"CV_32F", 4}},
        {"CV_64F", {"CV_64F", 8}},
    };

    // constructors
    OpenCVMatSerializer::OpenCVMatSerializer(const typenavigator::OpenCVMatNavigatorPtr& n) :
        detail::NDArraySerializerBase<typenavigator::OpenCVMatNavigator, OpenCVMatSerializer>("cv::Mat",
                simox::meta::get_type_name(typeid(data::AronNDArray)),
                simox::meta::get_type_name(typeid(type::AronOpenCVMat)), n)
    {
        ARMARX_CHECK_NOT_NULL(typenavigator);
    }

    std::pair<CppBlockPtr, std::string> OpenCVMatSerializer::getDimensionsFromAccessor(const std::string& accessor) const
    {
        CppBlockPtr b = CppBlockPtr(new CppBlock());

        std::string escaped_accessor = EscapeAccessor(accessor);
        std::string accessor_dimensions = escaped_accessor + DIMENSION_ACCESSOR;
        std::string accessor_iterator = escaped_accessor + ITERATOR_ACCESSOR;

        b->addLine("std::vector<int> " + accessor_dimensions + ";");
        b->addLine("for (int " + accessor_iterator + " = 0; " + accessor_iterator + " < " + accessor + nextEl() + "dims; ++" + accessor_iterator + ")");
        CppBlockPtr b2 = CppBlockPtr(new CppBlock());
        b2->addLine(accessor_dimensions + ".push_back(" + accessor + nextEl() + "size[" + accessor_iterator + "]);");
        b->addBlock(b2);
        return {b, accessor_dimensions};
    }

    CppBlockPtr OpenCVMatSerializer::getResetSoftBlock(const std::string& accessor) const
    {
        CppBlockPtr block_if_data = CppBlockPtr(new CppBlock());

        auto [get_dim_block, accessor_dimensions] = getDimensionsFromAccessor(accessor);
        block_if_data->appendBlock(get_dim_block);
        block_if_data->addLine("if (!" + accessor_dimensions + ".empty())");
        CppBlockPtr b2 = CppBlockPtr(new CppBlock());
        b2->addLine(accessor + " = " + getFullCppTypename() + "(" + accessor_dimensions + ", " + accessor + nextEl() + "type());");
        block_if_data->addBlock(b2);

        return ResolveMaybeResetSoftBlock(accessor, block_if_data, typenavigator);
    }

    CppBlockPtr OpenCVMatSerializer::getWriteTypeBlock(const std::string& accessor) const
    {
        CppBlockPtr b = CppBlockPtr(new CppBlock());
        b->addLine("w.writeOpenCVMat({" + MAYBE_TO_STRING(typenavigator->getMaybe()) + "}); // of " + accessor);
        return b;
    }

    CppBlockPtr OpenCVMatSerializer::getWriteBlock(const std::string& accessor) const
    {
        CppBlockPtr block_if_data = CppBlockPtr(new CppBlock());

        std::string escaped_accessor = EscapeAccessor(accessor);

        auto [get_dim_block, accessor_dimensions] = getDimensionsFromAccessor(accessor);
        block_if_data->appendBlock(get_dim_block);
        block_if_data->addLine(accessor_dimensions + ".push_back(" + accessor + nextEl() + "elemSize());");
        block_if_data->addLine("w.writeNDArray(" + accessor_dimensions + ", std::to_string(" + accessor + nextEl() + "type()), reinterpret_cast<const unsigned char*>(" + accessor + nextEl() + "data)); // of " + accessor);

        return ResolveMaybeWriteBlock(accessor, block_if_data, typenavigator);
    }

    CppBlockPtr OpenCVMatSerializer::getReadBlock(const std::string& accessor) const
    {
        CppBlockPtr block_if_data = CppBlockPtr(new CppBlock());
        std::string escaped_accessor = EscapeAccessor(accessor);
        std::string read_start_result_accessor = escaped_accessor + READ_START_RETURN_TYPE_ACCESSOR;

        block_if_data->addLine("if (!" + read_start_result_accessor + ".dims.empty())");
        CppBlockPtr b2 = CppBlockPtr(new CppBlock());
        b2->addLine(accessor + " = " + getCoreCppTypename() + "(std::vector<int>({" + read_start_result_accessor + ".dims.begin(), std::prev(" + read_start_result_accessor + ".dims.end())}), std::stoi(" + read_start_result_accessor + ".type));");
        b2->addLine("r.readEndNDArray(reinterpret_cast<unsigned char*>(" + accessor + nextEl() + "data)); // of " + accessor);
        block_if_data->addBlock(b2);

        return ResolveMaybeReadBlock(accessor, "r.readStartNDArray()", block_if_data, typenavigator, true);
    }

    CppBlockPtr OpenCVMatSerializer::getEqualsBlock(const std::string& accessor, const std::string& otherInstanceAccessor) const
    {
        CppBlockPtr block_if_data = CppBlockPtr(new CppBlock());
        block_if_data->addLine("if (cv::countNonZero(" + accessor + " != " + otherInstanceAccessor + ") != 0)");
        block_if_data->addLineAsBlock("return false;");
        return ResolveMaybeEqualsBlock(accessor, otherInstanceAccessor, block_if_data, typenavigator);
    }
}

