/*
* This file is part of ArmarX.
*
* ArmarX is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License version 2 as
* published by the Free Software Foundation.
*
* ArmarX is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program. If not, see <http://www.gnu.org/licenses/>.
*
* @author     Fabian Peller (fabian dot peller at kit dot edu)
* @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
*             GNU General Public License
*/

#pragma once

// STD/STL
#include <memory>
#include <string>

// Base Class
#include <RobotAPI/libraries/aron/core/io/typeIO/reader/ReaderToken.h>

// ArmarX
#include <RobotAPI/libraries/aron/core/Config.h>
#include <RobotAPI/libraries/aron/core/navigator/type/AllNavigators.h>

namespace armarx::aron::typeIO::reader
{
    class NavigatorReaderToken;
    typedef std::shared_ptr<NavigatorReaderToken> NavigatorReaderTokenPtr;

    class NavigatorReaderToken :
        virtual public ReaderTokenInterface<typenavigator::NavigatorPtr>
    {
    public:
        using PointerType = NavigatorReaderTokenPtr;

    public:
        // constructors
        NavigatorReaderToken() = delete;
        NavigatorReaderToken(const type::Descriptor desc, const typenavigator::NavigatorPtr& type)
        {
            descriptor = desc;
            element = type;

            switch (descriptor)
            {
                case type::Descriptor::eObject:
                {
                    for (const auto& [key, _] : typenavigator::ObjectNavigator::DynamicCastAndCheck(element)->getMemberTypes())
                    {
                        allMemberNames.push_back(key);
                    }
                    childrenSize = type->childrenSize();
                    break;
                }
                case type::Descriptor::eDict:
                {
                    typenavigator::DictNavigator::DynamicCastAndCheck(element);
                    childrenSize = 1;
                    break;
                }
                case type::Descriptor::eTuple:
                {
                    typenavigator::TupleNavigator::DynamicCastAndCheck(element);
                    childrenSize = type->childrenSize();
                    break;
                }
                case type::Descriptor::eList:
                {
                    typenavigator::ListNavigator::DynamicCastAndCheck(element);
                    childrenSize = 1;
                    break;
                }
                case type::Descriptor::ePair:
                {
                    typenavigator::PairNavigator::DynamicCastAndCheck(element);
                    childrenSize = 2;
                    break;
                }
                default:
                    throw error::DescriptorNotValidException("NavigatorReaderToken", "NavigatorReaderToken", "Received an invalid instance for a reader token.", desc);
            }
        }

        type::Descriptor getTypeOfNext() const override
        {
            const typenavigator::NavigatorPtr next = getNextElement();
            return next->getDescriptor();
        }

        typenavigator::NavigatorPtr getNextElement() const override
        {
            assertIndexInBounds();

            switch (descriptor)
            {
                case type::Descriptor::eObject:
                {
                    typenavigator::ObjectNavigatorPtr casted = typenavigator::ObjectNavigator::DynamicCastAndCheck(element);
                    return casted->getMemberType(getCurrentKey());
                }
                case type::Descriptor::eDict:
                {
                    typenavigator::DictNavigatorPtr casted = typenavigator::DictNavigator::DynamicCastAndCheck(element);
                    return casted->getAcceptedType();
                }
                case type::Descriptor::eTuple:
                {
                    typenavigator::TupleNavigatorPtr casted = typenavigator::TupleNavigator::DynamicCastAndCheck(element);
                    return casted->getAcceptedType(currentIndex);
                }
                case type::Descriptor::eList:
                {
                    typenavigator::ListNavigatorPtr casted = typenavigator::ListNavigator::DynamicCastAndCheck(element);
                    return casted->getAcceptedType();
                }
                case type::Descriptor::ePair:
                {
                    typenavigator::PairNavigatorPtr casted = typenavigator::PairNavigator::DynamicCastAndCheck(element);
                    if (currentIndex == 0)
                    {
                        return casted->getFirstAcceptedType();
                    }
                    return casted->getSecondAcceptedType();
                }
                default:
                    throw error::DescriptorNotValidException("NavigatorReaderToken", "getNextElement", "Could not resove a type of a navigator. Allowed are only container types due to performance", descriptor);
            }
        }
    };
}
