/*
* This file is part of ArmarX.
*
* ArmarX is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License version 2 as
* published by the Free Software Foundation.
*
* ArmarX is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program. If not, see <http://www.gnu.org/licenses/>.
*
* @author     Fabian Peller (fabian dot peller at kit dot edu)
* @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
*             GNU General Public License
*/

#pragma once

// STD/STL
#include <memory>
#include <string>

// Simox
#include <SimoxUtility/json.h>

// Base Class
#include <RobotAPI/libraries/aron/core/io/typeIO/reader/ReaderToken.h>

// ArmarX
#include <RobotAPI/libraries/aron/core/Config.h>
#include <RobotAPI/libraries/aron/core/Resolver.h>

namespace armarx::aron::typeIO::reader
{
    class NlohmannJSONReaderToken;
    typedef std::shared_ptr<NlohmannJSONReaderToken> NlohmannJSONReaderTokenPtr;

    class NlohmannJSONReaderToken :
        virtual public ReaderTokenInterface<nlohmann::json>
    {
    public:
        using PointerType = NlohmannJSONReaderTokenPtr;

    public:
        // constructors
        NlohmannJSONReaderToken() = delete;
        NlohmannJSONReaderToken(const type::Descriptor desc, const nlohmann::json& type)
        {
            descriptor = desc;
            element = type;

            switch (descriptor)
            {
                case type::Descriptor::eObject:
                {
                    for (auto& [key, val] : type.get<nlohmann::json::object_t>())
                    {
                        if (key == io::Data::READER_WRITER_OBJECT_NAME_SLUG)
                        {
                            continue;
                        }

                        allMemberNames.push_back(key);
                    }
                    childrenSize = allMemberNames.size();
                    elementName = element[io::Data::READER_WRITER_OBJECT_NAME_SLUG];
                    break;
                }
                case type::Descriptor::eDict:
                {
                    childrenSize = 1;
                    break;
                }
                case type::Descriptor::eTuple:
                {
                    childrenSize = element.size();
                    break;
                }
                case type::Descriptor::eList:
                {
                    childrenSize = 1;
                    break;
                }
                case type::Descriptor::ePair:
                {
                    childrenSize = 2;
                    break;
                }
                default:
                    throw error::DescriptorNotValidException("NlohmannJSONReaderToken", "NlohmannJSONReaderToken", "Received an invalid instance for a reader token.", desc);
            }
        }

        type::Descriptor getTypeOfNext() const override
        {
            const nlohmann::json next = getNextElement();

            if (next.is_object())
            {
                if (next.find(io::Data::READER_WRITER_DICT_ACCEPTED_TYPE_SLUG) != next.end())
                {
                    return type::Descriptor::eDict;
                }
                if (next.find(io::Data::READER_WRITER_LIST_ACCEPTED_TYPE_SLUG) != next.end())
                {
                    return type::Descriptor::eList;
                }
                if (next.find(io::Data::READER_WRITER_INT_ENUM_NAME_SLUG) != next.end())
                {
                    return type::Descriptor::eIntEnum;
                }
                if (next.find(io::Data::READER_WRITER_NDARRAY_NAME_SLUG) != next.end())
                {
                    if (next[io::Data::READER_WRITER_NDARRAY_NAME_SLUG] == "EigenMatrix")
                    {
                        return type::Descriptor::eEigenMatrix;
                    }
                    if (next[io::Data::READER_WRITER_NDARRAY_NAME_SLUG] == "EigenQuaternion")
                    {
                        return type::Descriptor::eEigenQuaternion;
                    }
                    if (next[io::Data::READER_WRITER_NDARRAY_NAME_SLUG] == "IVTCByteImage")
                    {
                        return type::Descriptor::eIVTCByteImage;
                    }
                    if (next[io::Data::READER_WRITER_NDARRAY_NAME_SLUG] == "OpenCVMat")
                    {
                        return type::Descriptor::eOpenCVMat;
                    }
                    if (next[io::Data::READER_WRITER_NDARRAY_NAME_SLUG] == "PCLPointCloud")
                    {
                        return type::Descriptor::ePCLPointCloud;
                    }
                    if (next[io::Data::READER_WRITER_NDARRAY_NAME_SLUG] == "Position")
                    {
                        return type::Descriptor::ePosition;
                    }
                    if (next[io::Data::READER_WRITER_NDARRAY_NAME_SLUG] == "Orientation")
                    {
                        return type::Descriptor::eOrientation;
                    }
                    if (next[io::Data::READER_WRITER_NDARRAY_NAME_SLUG] == "Pose")
                    {
                        return type::Descriptor::ePose;
                    }
                    throw error::AronException("NlohmannJSONReaderToken", "getTypeOfNextElement", "Could not determine the type of an nlohmann::json object. Found indicator of NDArray but could not resolve real type. Found JSON: " + next[io::Data::READER_WRITER_NDARRAY_NAME_SLUG].dump(2));
                }
                return type::Descriptor::eObject;
            }
            if (next.is_array())
            {
                return type::Descriptor::eTuple;
            }
            if (next.is_string())
            {
                if (next == io::Data::READER_WRITER_INT_TYPENAME_SLUG)
                {
                    return type::Descriptor::eInt;
                }
                if (next == io::Data::READER_WRITER_LONG_TYPENAME_SLUG)
                {
                    return type::Descriptor::eLong;
                }
                if (next == io::Data::READER_WRITER_FLOAT_TYPENAME_SLUG)
                {
                    return type::Descriptor::eFloat;
                }
                if (next == io::Data::READER_WRITER_DOUBLE_TYPENAME_SLUG)
                {
                    return type::Descriptor::eDouble;
                }
                if (next == io::Data::READER_WRITER_STRING_TYPENAME_SLUG)
                {
                    return type::Descriptor::eString;
                }
                if (next == io::Data::READER_WRITER_BOOL_TYPENAME_SLUG)
                {
                    return type::Descriptor::eBool;
                }
                if (next == io::Data::READER_WRITER_TIME_TYPENAME_SLUG)
                {
                    return type::Descriptor::eTime;
                }
            }
            throw error::AronException("NlohmannJSONReaderToken", "getTypeOfNextElement", "Could not determine the type of an nlohmann::json object. Could not convert to type::Descriptor enum. Found JSON: " + next.dump(2));
        }

        nlohmann::json getNextElement() const override
        {
            assertIndexInBounds();

            switch (descriptor)
            {
                case type::Descriptor::eObject:
                {
                    return element[getCurrentKey()];
                }
                case type::Descriptor::eDict:
                {
                    return element[io::Data::READER_WRITER_DICT_ACCEPTED_TYPE_SLUG];
                }
                case type::Descriptor::eTuple:
                {
                    return element[currentIndex];
                }
                case type::Descriptor::eList:
                {
                    return element[io::Data::READER_WRITER_LIST_ACCEPTED_TYPE_SLUG];
                }
                default:
                    throw error::DescriptorNotValidException("NlohmannJSONReaderToken", "getNextElement", "Could not resove a type of a navigator. Allowed are only container types due to performance", descriptor);
            }
        }
    };
}
